# -*- coding: utf-8 -*-
# author: Ethosa
from typing import Dict, Any, Union, Optional, Literal, List


class UnsupportedResponseFormat(Exception):
    pass


class Uploader:
    def __init__(self, vk):
        """Initializes new Uploader object.
        """
        self.session = vk.session
        self.call_method = vk.call_method
        self.logger = vk.logger

    def album_photo(
            self,
            files: Union[str, List[str]],
            album_id: int,
            group_id: Optional[int] = None,
            caption: str = ""
    ) -> Dict[str, Any]:
        """Uploads photo in album

        :param files: file path, file paths, file as bytes or files as bytes
        :param album_id: album ID.
        :param group_id: group ID.
        :param caption: photo caption
        """
        data = {
            "album_id": album_id
        }
        if group_id:
            data["group_id"] = group_id

        response = self._upload_files(data, files, "photos.getUploadServer")

        data["caption"] = caption
        data["server"] = response["server"]
        data["hash"] = response["hash"]
        data["photos_list"] = response["photos_list"]

        return self.call_method("photos.save", data)

    def audio(
            self,
            files: Union[str, List[str]],
            artist: str = "",
            title: str = ""
    ) -> Dict[str, Any]:
        """Uploads audio file

        :param files: file path, file paths, file as bytes or files as bytes
        :param artist: songwriter. The default is taken from ID3 tags.
        :param title: name of the composition. The default is taken from ID3 tags.
        """
        response = self._upload_files({}, files, "audio.getUploadServer")
        data = {
            "server": response["server"],
            "audio": response["audio"],
            "hash": response["hash"]
        }

        for field_name, value in (("artist", artist), ("title", title)):
            if value:
                data[field_name] = value

        return self.call_method("audio.save", data)

    def chat_photo(
            self,
            files: Union[str, List[str]],
            chat_id: int,
            crop_x: Optional[int] = None,
            crop_y: Optional[int] = None,
            crop_width: Optional[int] = None
    ) -> Dict[str, Any]:
        """Uploads chat photo

        :param files: file path, file paths, file as bytes or files as bytes
        :param chat_id: id of the conversation for which you want to upload a photo
        :param crop_x: x coordinate for cropping the photo (upper right corner).
        :param crop_y: y coordinate for cropping the photo (upper right corner).
        :param crop_width: Width of the photo after cropping in px.
        """
        data = {"chat_id": chat_id}
        if crop_x:
            data["crop_x"] = crop_x
        if crop_y:
            data["crop_y"] = crop_y
        if crop_width:
            data["crop_width"] = crop_width

        response = self._upload_files(
            data, files, "photos.getChatUploadServer")

        data = {"file": response["response"]}

        return self.call_method("messages.setChatPhoto", data)

    def cover_photo(
            self,
            files: Union[str, List[str]],
            group_id: int,
            crop_x: int = 0,
            crop_y: int = 0,
            crop_x2: int = 795,
            crop_y2: int = 200
    ) -> Dict[str, Any]:
        """Updates group cover photo

        :param files: file path, file paths, file as bytes or files as bytes
        :param group_id: community id.
        :param crop_x: X coordinate of the upper left corner to crop the image.
        :param crop_y: Y coordinate of the upper left corner to crop the image.
        :param crop_x2: X coordinate of the lower right corner to crop the image.
        :param crop_y2: Y coordinate of the lower right corner to crop the image.
        """
        data = {
            "group_id": group_id,
            "crop_x": crop_x,
            "crop_y": crop_y,
            "crop_x2": crop_x2,
            "crop_y2": crop_y2
        }

        response = self._upload_files(
            data, files, "photos.getOwnerCoverPhotoUploadServer")
        data = {
            "hash": response["hash"],
            "photo": response["photo"]
        }

        return self.call_method("photos.saveOwnerCoverPhoto", data)

    def document(
            self,
            files: Union[str, List[str]],
            group_id: Optional[int] = None,
            title: str = "",
            tags: str = "",
            return_tags: int = 0,
            is_wall: bool = False
    ) -> Dict[str, Any]:
        """Uploads document

        :param files: file path, file paths, file as bytes or files as bytes
        :param group_id: community identifier (if you need to upload a document to the list of community documents).
        :param title: document's name.
        :param tags: tags for search.
        :param is_wall: upload document in wall?
        """
        data = {}
        if group_id:
            data["group_id"] = group_id

        response = self._upload_files(
            data, files,
            "docs.get%sUploadServer" % ("Wall" if is_wall else "")
        )
        data = {
            "file": response["file"],
            "title": title,
            "tags": tags,
            "return_tags": return_tags
        }

        return self.call_method("docs.save", data)

    def document_message(
            self,
            files,
            peer_id,
            doc_type="doc",
            title="",
            tags="",
            return_tags=0
    ) -> Dict[str, Any]:
        """Uploads document in message.

        :param files: file path, file paths, file as bytes or files as bytes
        :param peer_id:  destination identifier.
        :param doc_type:  type of document. Possible values: doc, audio_message
        :param title:  document's name.
        :param tags:  tags for search.
        """
        data = {
            "peer_id": peer_id,
            "type": doc_type
        }

        response = self._upload_files(
            data, files, "docs.getMessagesUploadServer"
        )
        self.logger.debug(response)
        data = {
            "file": response["file"],
            "title": title,
            "tags": tags,
            "return_tags": return_tags
        }

        return self.call_method("docs.save", data)

    @staticmethod
    def format(
            response,
            formtype: Literal[
                'photo', 'video',
                'audio', 'doc', 'graffiti'
            ] = 'photo'
    ) -> str:
        """response formatting

        :param response: response after object saved
        """
        if isinstance(response, dict):
            response = response.get("response", response)
        if isinstance(response, dict):
            response = response.get("type", response)

        if isinstance(response, dict):
            if "owner_id" in response and "id" in response:
                return (
                    "%s%s_%s" % (formtype, response["owner_id"], response["id"])
                )
        elif isinstance(response, list):
            objs = []
            for obj in response:
                if "owner_id" in obj and "id" in obj:
                    objs.append(
                        "%s%s_%s" % (formtype, obj["owner_id"], obj["id"])
                    )
            return ",".join(objs)
        raise UnsupportedResponseFormat

    def message_photo(
            self,
            files: Union[str, List[str]],
            peer_id: int,
            group_id: int = 0
    ) -> Dict[str, Any]:
        """Uploads photo in message

        :param files: file path, file paths, file as bytes or files as bytes
        :param peer_id: destination identifier (for uploading photos in community posts).
        """
        data = {"peer_id": peer_id} if group_id else {}

        response = self._upload_files(
            data, files, "photos.getMessagesUploadServer"
        )

        data["server"] = response["server"]
        data["hash"] = response["hash"]
        data["photo"] = response["photo"]

        return self.call_method("photos.saveMessagesPhoto", data)

    def market_photo(
            self,
            files: Union[str, List[str]],
            group_id: int,
            main_photo: Optional[int] = None,
            crop_x: Optional[int] = None,
            crop_y: Optional[int] = None,
            crop_width: Optional[int] = None
    ) -> Dict[str, Any]:
        """upload product photo

        Allowed formats: JPG, PNG, GIF.
        Limitations: the minimum photo size is 400x400px,
        the sum of the height and width is no more than 14000px,
        the file is no more than 50 MB in size, and the aspect ratio is at least 1:20.

        :param files: file path, file paths, file as bytes or files as bytes
        :param group_id Community id for which you want to upload a product photo.
        :param main_photo: whether the photo is the cover
                           of the product (1 - cover photo, 0 - additional photo)
        :param crop_x: x coordinate for cropping the photo (upper right corner).
        :param crop_y: y coordinate for cropping the photo (upper right corner).
        :param crop_width: Width of the photo after cropping in px.
        """
        data = {"group_id": group_id}
        if crop_x:
            data["crop_x"] = crop_x
        if crop_y:
            data["crop_y"] = crop_y
        if crop_width:
            data["crop_width"] = crop_width
        if main_photo:
            data["main_photo"] = main_photo

        response = self._upload_files(
            data, files, "photos.getMarketUploadServer"
        )

        data = {
            "group_id": group_id,
            "server": response["server"],
            "hash": response["hash"],
            "photo": response["photo"],
            "crop_data": response["crop_data"],
            "crop_hash": response["crop_hash"]
        }

        return self.call_method("photos.saveMarketPhoto", data)

    def market_album_photo(
            self,
            files: Union[str, List[str]],
            group_id: int
    ) -> Dict[str, Any]:
        """Uploading photos for a selection of goods

        Allowed formats: JPG, PNG, GIF.
        Limitations: the minimum photo size is 400x400px,
        the sum of the height and width is no more than 14000px,
        the file is no more than 50 MB in size, and the aspect ratio is at least 1:20.

        :param files: file path, file paths, file as bytes or files as bytes
        :param group_id: Community id for which you want to upload a product photo.
        """
        data = {"group_id": group_id}

        response = self._upload_files(
            data, files, "photos.getMarketAlbumUploadServer"
        )

        data["server"] = response["server"]
        data["hash"] = response["hash"]
        data["photo"] = response["photo"]

        return self.call_method("photos.saveMarketAlbumPhoto", data)

    def profile_photo(self, files, owner_id=None):
        """update profile photo

        :param files: file path, file paths, file as bytes or files as bytes
        :param owner_id: id of the community or current user. (default id of current user)
        """
        data = {}
        if owner_id:
            data["owner_id"] = owner_id

        response = self._upload_files(
            data, files, "photos.getOwnerPhotoUploadServer"
        )
        del data["owner_id"]

        data["server"] = response["server"]
        data["hash"] = response["hash"]
        data["photo"] = response["photo"]

        return self.call_method("photos.saveOwnerPhoto", data)

    def video(
            self,
            files: Union[str, List[str]],
            album_id: int,
            name: str = "",
            description: str = "",
            is_private: int = 0,
            wallpost: int = 0,
            link: str = "",
            group_id: int = 0,
            privacy_view: str = "",
            privacy_comment: str = "",
            no_comments: int = 0,
            repeat: int = 0,
            compression: int = 0
    ) -> Dict[str, Any]:
        """Uploads video

        :param files: file path, file paths, file as bytes or files as bytes
        :param album_id: id of the album into which the video file will be uploaded.
        :param name: name of the video file.
        :param description: description of the video file.
        :param is_private: 1 is indicated if the video is uploaded for sending by personal message.
        :param wallpost: whether after saving it is required to publish a video recording on the wall
        :param link: url for embedding video from an external site. e.g. Youtube.
        :param group_id: id of the community where the video will be saved.
        :param privacy_view: privacy settings for viewing video in a special format.
        :param privacy_comment: privacy settings for commenting on a video in a special format.
        :param no_comments: 1 - close comments (for videos from communities).
        :param repeat: looping video playback.
        """
        data = {
            "name": name,
            "description": description,
            "is_private": is_private,
            "wallpost": wallpost,
            "link": link,
            "album_id": album_id,
            "privacy_view": privacy_view,
            "privacy_comment": privacy_comment,
            "no_comments": no_comments,
            "repeat": repeat,
            "compression": compression
        }
        if group_id:
            data["group_id"] = group_id

        upload_url = self.call_method(
            "video.save", data)["response"]["upload_url"]
        response = []

        if isinstance(files, str):
            files = [files]

        for _, file in enumerate(files):
            response.append(
                self.session.post(
                    upload_url, files={"file": open(file, "rb")}
                ).json()
            )

        return response

    def _upload_files(
            self,
            data: Dict[str, Any],
            files: Union[str, List[str]],
            method: str
    ) -> Dict[str, Any]:
        upload_url = self.call_method(method, data)["response"]["upload_url"]
        uplfiles = {}

        if not isinstance(files, list):
            files = [files]

        if len(files) > 1:
            for index, file in enumerate(files):
                if isinstance(file, str):
                    uplfiles["file%d" % (index+1)] = open(file, "rb")
                elif isinstance(file, bytes):
                    uplfiles["file%d" % (index+1)] = file
        else:
            if isinstance(files[0], str):
                uplfiles["file"] = open(files[0], "rb")
            elif isinstance(files[0], bytes):
                uplfiles["file"] = files[0]

        return self.session.post(upload_url, files=uplfiles).json()

    def wall_photo(
            self,
            files: Union[str, List[str]],
            group_id: Optional[int] = None,
            user_id: Optional[int] = None,
            caption: str =""
    ) -> Dict[str, Any]:
        """upload photo in wall post

        :param files: file path, file paths, file as bytes or files as bytes
        :param group_id: id of the community on whose wall you want to upload the photo (without a minus sign).
        :param user_id: id of the user whose wall you want to save the photo on.
        :param caption: photo description text (maximum 2048 characters)
        """
        data = {}
        if group_id:
            data["group_id"] = group_id

        response = self._upload_files(data, files, "photos.getWallUploadServer")

        if user_id:
            data["user_id"] = user_id
        data["caption"] = caption
        data["server"] = response["server"]
        data["hash"] = response["hash"]
        data["photo"] = response["photo"]

        return self.call_method("photos.saveWallPhoto", data)



# -*- coding: utf-8 -*-
# author: Ethosa
from typing import Union, Optional, Dict, Any, NoReturn
from logging import getLogger, StreamHandler, Formatter

from requests import Session

from .longpoll import LongPoll
from .uploader import Uploader
from .vk_auth import VkAuthManager
from .vks import VkScript
from ..StartThread import StartThread


class Vk(object):
    def __init__(
            self,
            token: str = "",
            group_id: Union[str, int] = "",
            login: str = "",
            password: str = "",
            api: str = "5.103",
            debug: bool = False
    ):
        """auth in VK

        :param token: access_token
        :param group_id: group id if you want to log in through the group
        :param login: login. used for authorization through the user
        :param password: password. used for authorization through the user
        :param api: api version
        :param debug: debug log
        """
        self.session = Session()

        # Parses vk.com, if login and password are not empty.
        if login and password:
            self.auth = VkAuthManager()
            self.auth.login(login, password)
            token = self.auth.get_token()

        self.v = api
        self.token = token
        self.group_id = group_id

        self.method = ""
        self.events = {}

        # Debug settings.
        self.debug = 50
        if debug:
            if isinstance(debug, int):
                self.debug = debug
            else:
                self.debug = 10

        # Logger initialize.
        self.logger = getLogger("saya")
        self.logger.setLevel(self.debug)
        handler = StreamHandler()
        handler.setFormatter(
            Formatter("[%(levelname)s] %(name)s: %(asctime)s — %(message)s")
        )
        self.logger.addHandler(handler)

        self.uploader = Uploader(self)
        self.longpoll = LongPoll(self)
        self.vks = VkScript()  # for pyexecute method.

    def call_method(
            self,
            method: str,
            data: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """call to any method in VK api

        :param method: method name e.g. "messages.send", "wall.post"
        :param data: data to send
        """
        if not data:
            data = {}
        data["v"] = self.v
        data["access_token"] = self.token
        response = self.session.post(
            "https://api.vk.com/method/%s" % method,
            data=data, timeout=30
        ).json()

        # Logging.
        if "error" in response:
            self.logger.error(
                'Error [%s] in called method "%s": %s' % (
                    response["error"]["error_code"], method,
                    response["error"]["error_msg"]
                )
            )
        else:
            self.logger.debug('Successfully called method "%s"' % method)
        return response

    def execute(
            self,
            code: str
    ) -> Dict[str, Any]:
        """Calls an execute VK API method

        :param code: VKScript code.
        """
        return self.call_method("execute", {"code": code})

    def pyexecute(
            self,
            code
    ) -> Dict[str, Any]:
        """Calls an execute VK API method

        :param code: Python code.
        """
        return self.execute(self.vks.translate(code))

    def start_listen(self) -> NoReturn:
        """Starts receiving events from the server.
        """
        for event in self.longpoll.listen(True):
            if "type" in event:
                if event["type"] in self.events:
                    self.events[event["type"]](event)
                elif event["type"] in dir(self):
                    getattr(self, event["type"])(event)
            else:
                self.logger.warning('Unknown event passed: "%s"' % event)

    def __getattr__(
            self,
            attr: str
    ) -> Any:
        """A convenient alternative for the call_method method.

        :param attr: method name e.g. messages.send, wall.post
        """
        if attr.startswith("on_"):  # e.g. on_message_new
            attr = attr[3:]

            def decorator(obj):
                def listen(_f):
                    for event in self.longpoll.listen(True):
                        if event["type"] == attr:
                            obj(event)

                if isinstance(obj, callable):
                    StartThread(listen, obj).start()
                else:
                    # obj should be boolean
                    if obj:
                        def _decorator(call):
                            StartThread(listen, call).start()
                    else:
                        def _decorator(call):
                            self.events[attr] = call
                            return call
                    return _decorator

            return decorator
        elif self.method:
            method = "%s.%s" % (self.method, attr)
            self.method = ""
            return lambda **data: self.call_method(method, data)
        else:
            self.method = attr
            return self



# -*- coding: utf-8 -*-
# author: Ethosa
from typing import NoReturn, Optional, List, Dict, Any

class TemplateElement(dict):
    def __init__(
            self,
            title: str = "",
            description: str = "",
            photo_id: str = "",
            buttons: Optional[List[Dict[str, Any]]] = None,
            action: str = "open_link",
            link: str = "https://vk.com"
    ):
        """Initializes a new TemplateElement object.

        Keyword Arguments:
            :param title: template element title
            :param description: template element description
            :param photo_id:  Id of the picture, which must attach in the template element.
            :param buttons:  Buttons in TemplateElement.
            :param action:  Action of TemplateElement, can be "open_link" or "open_photo".
            :param link:  link for "open_link" action.
        """
        self["title"] = title
        self["description"] = description
        self["photo_id"] = photo_id
        self["buttons"] = buttons or []
        self["action"] = {"type": action}
        if action == "open_link":
            self["action"]["link"] = link

    def add(self, button) -> NoReturn:
        """Adds a new Button in TemplateElement.
        """
        if len(self["buttons"]) < 3:
            self["buttons"].append(button)



# -*- coding: utf-8 -*-
# author: Ethosa

from .vk import Vk
from .longpoll import LongPoll
from .vk_auth import VkAuthManager
from .event import event
from .vks import VkScript
from .uploader import Uploader
from .button import create_button
from .keyboard import Keyboard
from .template_element import TemplateElement
from .template import Template
from .StreamingAPI import StreamingAPI



# -*- coding: utf-8 -*-
# author: Ethosa
from typing import Dict, Any


def create_button(
        button_type: str = "text",
        color: str = "primary",
        **kwargs
) -> Dict[str, Any]:
    """Creates a new Button object.

    :param button_type: button type.
    :param color: button color. Can be "primary", "secondary", "positive" or "negative".
    """
    obj = {}
    if button_type == "text":
        obj["color"] = color
    obj["action"] = kwargs
    obj["action"]["type"] = button_type
    return obj



# -*- coding: utf-8 -*-
# author: Ethosa
from typing import NoReturn, Dict, Any
from json import dumps


class Template(dict):
    # noinspection PyMissingConstructor
    def __init__(
            self,
            ttype: str = "carousel"
    ):
        """Creates a Template object.

        :param ttype: template type.
        """
        self["type"] = ttype
        self["elements"] = []

    def add(
            self,
            element: Dict[str, Any]
    ) -> NoReturn:
        """Adds new element in template.

        :param element: template element
        """
        if len(self["elements"]) < 10:
            self["elements"].append(element)

    def compile(self) -> str:
        """Compiles Template for sendng in the message.
        """
        return dumps(self)
